@@ -195,7 +195,7 @@ class Gateway(server.Gateway):
         """Cast status to bytes representation of current Python version.
 
         According to :pep:`3333`, when using Python 3, the response status
-        and headers must be bytes masquerading as unicode; that is, they
+        and headers must be bytes masquerading as Unicode; that is, they
         must be of type "str" but are restricted to code points in the
         "latin-1" set.
         """
@@ -300,7 +300,7 @@ class Gateway_10(Gateway):
 
         # Request headers
         env.update(
-            ('HTTP_' + bton(k).upper().replace('-', '_'), bton(v))
+            ('HTTP_{!s}'.format(bton(k).upper().replace('-', '_')), bton(v))
             for k, v in req.inheaders.items()
         )
 
@@ -321,7 +321,7 @@ class Gateway_10(Gateway):
 class Gateway_u0(Gateway_10):
     """A Gateway class to interface HTTPServer with WSGI u.0.
 
-    WSGI u.0 is an experimental protocol, which uses unicode for keys
+    WSGI u.0 is an experimental protocol, which uses Unicode for keys
     and values in both Python 2 and Python 3.
     """
 
@@ -359,7 +359,7 @@ class Gateway_u0(Gateway_10):
     def _decode_value(item):
         k, v = item
         skip_keys = 'REQUEST_URI', 'wsgi.input'
-        if six.PY3 or not isinstance(v, bytes) or k in skip_keys:
+        if not six.PY2 or not isinstance(v, bytes) or k in skip_keys:
             return k, v
         return k, v.decode('ISO-8859-1')
 
@@ -409,7 +409,7 @@ class PathInfoDispatcher:
         path = environ['PATH_INFO'] or '/'
         for p, app in self.apps:
             # The apps list should be sorted by length, descending.
-            if path.startswith(p + '/') or path == p:
+            if path.startswith('{path!s}/'.format(path=p)) or path == p:
                 environ = environ.copy()
                 environ['SCRIPT_NAME'] = environ.get('SCRIPT_NAME', '') + p
                 environ['PATH_INFO'] = path[len(p):]
